from __future__ import annotations

import json
from typing import IO, TYPE_CHECKING, Any, Literal, cast, overload

import polars as pl
import polars.selectors as cs
from polars import DataFrame, Expr
from polars.api import register_dataframe_namespace
from polars.datatypes import N_INFER_DEFAULT
from pyogrio import write_arrow

from polars_st._lib import get_crs_from_code
from polars_st.casting import st
from polars_st.geoseries import GeoSeries
from polars_st.selectors import geom

if TYPE_CHECKING:
    from io import BytesIO
    from pathlib import Path

    import altair as alt
    import geopandas as gpd
    from altair.vegalite.v6.schema._config import MarkConfigKwds
    from lonboard import Map
    from lonboard.types.layer import (
        PathLayerKwargs,
        PolygonLayerKwargs,
        ScatterplotLayerKwargs,
    )
    from lonboard.types.map import MapKwargs
    from polars._typing import (
        FrameInitTypes,
        JoinStrategy,
        JoinValidation,
        Orientation,
        SchemaDefinition,
        SchemaDict,
    )
    from typing_extensions import Unpack

__all__ = [
    "GeoDataFrame",
    "GeoDataFrameNameSpace",
]


class GeoDataFrameMeta(type):
    def __instancecheck__(cls, instance: Any) -> bool:
        # The GeoDataFrame constructor doesn't return an instance of GeoDataFrame but an
        # instance of pl.DataFrame. This design decision is made because Polars doesn't
        # support subclassing of its code datatypes. In order to prevent misuse,
        # instance checks are forbidden.
        msg = "instance check on abstract class GeoDataFrame is not allowed"
        raise TypeError(msg)


class GeoDataFrame(DataFrame, metaclass=GeoDataFrameMeta):
    @property
    def st(self) -> GeoDataFrameNameSpace: ...

    def __new__(  # noqa: PYI034
        cls,
        data: FrameInitTypes | None = None,
        schema: SchemaDefinition | None = None,
        *,
        geometry_name: str = "geometry",
        geometry_format: Literal[
            "wkb",
            "wkt",
            "ewkt",
            "geojson",
            "shapely",
            "point",
            "multipoint",
            "linestring",
            "circularstring",
            "multilinestring",
            "polygon",
        ]
        | None = None,
        schema_overrides: SchemaDict | None = None,
        strict: bool = True,
        orient: Literal["col", "row"] | None = None,
        infer_schema_length: int | None = N_INFER_DEFAULT,
        nan_to_null: bool = False,
    ) -> GeoDataFrame:
        df = pl.DataFrame(
            data,
            schema,
            schema_overrides=schema_overrides,
            strict=strict,
            orient=orient,
            infer_schema_length=infer_schema_length,
            nan_to_null=nan_to_null,
        )
        if df.columns == ["column_0"]:
            df = df.rename({"column_0": geometry_name})
        if geometry_name not in df.columns:
            msg = f'geometry column "{geometry_name}" not found'
            raise ValueError(msg)

        geometry_column = df.get_column(geometry_name)
        df = df.with_columns(GeoSeries(geometry_column, geometry_format=geometry_format))
        return cast("GeoDataFrame", df)

    def __init__(
        self,
        data: FrameInitTypes | None = None,
        schema: SchemaDefinition | None = None,
        *,
        geometry_name: str = "geometry",
        geometry_format: Literal[
            "wkb",
            "wkt",
            "ewkt",
            "geojson",
            "shapely",
            "point",
            "multipoint",
            "linestring",
            "circularstring",
            "multilinestring",
            "polygon",
        ]
        | None = None,
        schema_overrides: SchemaDict | None = None,
        strict: bool = True,
        orient: Orientation | None = None,
        infer_schema_length: int | None = N_INFER_DEFAULT,
        nan_to_null: bool = False,
    ) -> None:
        """Create a new GeoDataFrame.

        `GeoDataFrame` is used as an alias for `pl.DataFrame` with type annotations added for
        the [`st`][polars_st.GeoDataFrame.st] namespace, and an overriden constructor which will
        parse the column identified by `geometry_name` (default `"geometry"`) into a
        [`GeoSeries`][polars_st.GeoSeries].

        See [`pl.DataFrame`](https://docs.pola.rs/api/python/stable/reference/dataframe/index.html)
        for parameters documentation.

        !!! note

            Because Polars doesn't support subclassing of their types, calling this constructor will
            **NOT** create an instance of `GeoDataFrame`, but an instance of `pl.DataFrame`.

            As a result, instance checks are not permitted on this class to prevent misuse:
            ```pycon
            >>> gdf = st.GeoDataFrame(["POINT(0 0)"])
            >>> type(gdf)
            <class 'polars.dataframe.frame.DataFrame'>
            >>> isinstance(gdf, st.GeoDataFrame)
            Traceback (most recent call last):
            ...
            TypeError: instance check on abstract class GeoDataFrame is not allowed
            ```


        Examples:
            >>> gdf = st.GeoDataFrame({
            ...     "geometry": [
            ...         "POINT(0 0)",
            ...         "POINT(1 2)",
            ...     ]
            ... })
            >>> gdf.schema
            Schema({'geometry': Binary})

            >>> gdf = st.GeoDataFrame(
            ...     {
            ...         "geom": [
            ...             '{"type": "Point", "coordinates": [0, 0]}',
            ...             '{"type": "Point", "coordinates": [1, 2]}',
            ...         ]
            ...     },
            ...     geometry_name="geom",
            ...     geometry_format="geojson",
            ... )
            >>> gdf.schema
            Schema({'geom': Binary})
        """
        ...


@register_dataframe_namespace("st")
class GeoDataFrameNameSpace:
    def __init__(self, df: DataFrame) -> None:
        self._df = cast("GeoDataFrame", df)

    def sjoin(
        self,
        other: DataFrame,
        on: str | Expr = "geometry",
        how: JoinStrategy = "inner",
        predicate: Literal[
            "intersects_bbox",
            "intersects",
            "within",
            "dwithin",
            "contains",
            "overlaps",
            "crosses",
            "touches",
            "covers",
            "covered_by",
            "contains_properly",
        ] = "intersects",
        distance: float | None = None,
        *,
        left_on: str | Expr | None = None,
        right_on: str | Expr | None = None,
        suffix: str = "_right",
        validate: JoinValidation = "m:m",
        coalesce: bool | None = None,
    ) -> GeoDataFrame:
        """Perform a spatial join operation with another DataFrame."""
        if not isinstance(other, DataFrame):
            msg = f"expected `other` join table to be a DataFrame, got {type(other).__name__!r}"
            raise TypeError(msg)

        return (
            self._df
            .lazy()
            .pipe(st)
            .sjoin(
                other=other.lazy(),
                left_on=left_on,
                right_on=right_on,
                on=on,
                how=how,
                predicate=predicate,
                distance=distance,
                suffix=suffix,
                validate=validate,
                coalesce=coalesce,
            )
            .collect()
            .pipe(lambda df: cast("GeoDataFrame", df))
        )

    def to_wkt(
        self,
        *geometry_columns: str,
        rounding_precision: int | None = 6,
        trim: bool = True,
        output_dimension: Literal[2, 3, 4] = 3,
        old_3d: bool = False,
    ) -> DataFrame:
        """Serialize the DataFrame geometry column as WKT.

        See [`GeoExprNameSpace.to_wkt`][polars_st.GeoExprNameSpace.to_wkt].
        """
        return self._df.with_columns(
            geom(*geometry_columns).st.to_wkt(
                rounding_precision,
                trim,
                output_dimension,
                old_3d,
            ),
        )

    def to_ewkt(
        self,
        *geometry_columns: str,
        rounding_precision: int | None = 6,
        trim: bool = True,
        output_dimension: Literal[2, 3, 4] = 3,
        old_3d: bool = False,
    ) -> DataFrame:
        """Serialize the DataFrame geometry column as EWKT.

        See [`GeoExprNameSpace.to_ewkt`][polars_st.GeoExprNameSpace.to_ewkt].
        """
        return self._df.with_columns(
            geom(*geometry_columns).st.to_ewkt(
                rounding_precision,
                trim,
                output_dimension,
                old_3d,
            ),
        )

    def to_wkb(
        self,
        *geometry_columns: str,
        output_dimension: Literal[2, 3, 4] = 3,
        byte_order: Literal[0, 1] | None = None,
        include_srid: bool = False,
    ) -> DataFrame:
        """Serialize the DataFrame geometry column as WKB.

        See [`GeoExprNameSpace.to_wkb`][polars_st.GeoExprNameSpace.to_wkb].
        """
        return self._df.with_columns(
            geom(*geometry_columns).st.to_wkb(
                output_dimension,
                byte_order,
                include_srid,
            ),
        )

    def to_geojson(self, *geometry_columns: str, indent: int | None = None) -> DataFrame:
        """Serialize the DataFrame geometry column as GeoJSON.

        See [`GeoExprNameSpace.to_geojson`][polars_st.GeoExprNameSpace.to_geojson].
        """
        return self._df.with_columns(geom(*geometry_columns).st.to_geojson(indent))

    def to_shapely(self, *geometry_columns: str) -> DataFrame:
        """Convert the DataFrame geometry column to a shapely representation.

        See [`GeoExprNameSpace.to_shapely`][polars_st.GeoExprNameSpace.to_shapely].
        """
        return self._df.with_columns(geom(*geometry_columns).st.to_shapely())

    def to_dict(self, *geometry_columns: str) -> DataFrame:
        """Convert the DataFrame geometry column to a GeoJSON-like Python [`dict`][] representation.

        See [`GeoExprNameSpace.to_dict`][polars_st.GeoExprNameSpace.to_dict].
        """
        return self._df.with_columns(geom(*geometry_columns).st.to_dict())

    def to_dicts(self, geometry_name: str = "geometry") -> list[dict[str, Any]]:
        """Convert every row to a Python [`dict`][] representation of a GeoJSON Feature.

        Examples:
            >>> gdf = st.GeoDataFrame({
            ...     "name": ["Alice", "Bob"],
            ...     "location": ["POINT(0 0)", "POINT(1 2)"],
            ... }, geometry_name="location")
            >>> dicts = gdf.st.to_dicts("location")
            >>> dicts[0]
            {'type': 'Feature', 'geometry': {'type': 'Point', 'coordinates': [0.0, 0.0]}, 'properties': {'name': 'Alice'}}
        """  # noqa: E501
        geometry_column_idx = self._df.get_column_index(geometry_name)
        property_columns_idx = [i for i in range(len(self._df.columns)) if i != geometry_column_idx]

        def to_geojson_dict(row: tuple) -> dict:
            return {
                "type": "Feature",
                "geometry": row[geometry_column_idx],
                "properties": {self._df.columns[i]: row[i] for i in property_columns_idx},
            }

        return [to_geojson_dict(row) for row in self.to_dict(geometry_name)._df.row_tuples()]  # noqa: SLF001

    def to_geopandas(
        self,
        *,
        geometry_name: str = "geometry",
        use_pyarrow_extension_array: bool = False,
        **kwargs: Any,
    ) -> gpd.GeoDataFrame:
        """Convert this DataFrame to a geopandas GeoDataFrame."""
        import geopandas as gpd

        return gpd.GeoDataFrame(
            self.to_shapely(geometry_name).to_pandas(
                use_pyarrow_extension_array=use_pyarrow_extension_array,
                **kwargs,
            ),
            geometry=geometry_name,
            crs=get_unique_crs_or_raise(self._df, geometry_name),
        )

    @property
    def __geo_interface__(self) -> dict:
        """Return a GeoJSON FeatureCollection [`dict`][] representation of the DataFrame.

        Examples:
            >>> gdf = st.GeoDataFrame({
            ...     "geometry": ["POINT(0 0)", "POINT(1 2)"],
            ...     "name": ["Alice", "Bob"]
            ... })
            >>> interface = gdf.st.__geo_interface__
            >>> pprint.pp(interface)
            {'type': 'FeatureCollection',
             'features': [{'type': 'Feature',
                           'geometry': {'type': 'Point', 'coordinates': [0.0, 0.0]},
                           'properties': {'name': 'Alice'}},
                          {'type': 'Feature',
                           'geometry': {'type': 'Point', 'coordinates': [1.0, 2.0]},
                           'properties': {'name': 'Bob'}}]}
        """
        return {
            "type": "FeatureCollection",
            "features": self.to_dicts(),
        }

    def write_file(
        self,
        path: str | BytesIO,
        layer: str | None = None,
        driver: str | None = None,
        geometry_name: str | None = "geometry",
        crs: str | None = None,
        encoding: str | None = None,
        append: bool = False,
        dataset_metadata: dict | None = None,
        layer_metadata: dict | None = None,
        metadata: dict | None = None,
        dataset_options: dict | None = None,
        layer_options: dict | None = None,
        **kwargs: dict[str, Any],
    ) -> None:
        """Write the GeoDataFrame to an OGR supported file format.

        Args:
            path:
                path to output file on writeable file system or an io.BytesIO object to
                allow writing to memory
                NOTE: support for writing to memory is limited to specific drivers.
            layer:
                layer name to create.  If writing to memory and layer name is not
                provided, it layer name will be set to a UUID4 value.
            driver:
                The OGR format driver used to write the vector file. By default attempts
                to infer driver from path.  Must be provided to write to memory.
                The available drivers can be listed by calling:
                ```py
                >>> import pyogrio
                >>> pyogrio.list_drivers() # doctest: +SKIP
                {..., 'GeoJSON': 'rw', 'GeoJSONSeq': 'rw',...}
                ```

            geometry_name:
                The name of the column in the input data that will be written as the
                geometry field, or `None` if the geometry should not be written.
            crs:
                WKT-encoded CRS of the geometries to be written. If not provided, the
                CRS will be infered from the geometries.
            encoding:
                Only used for the .dbf file of ESRI Shapefiles. If not specified,
                uses the default locale.
            append:
                If True, the data source specified by path already exists, and the
                driver supports appending to an existing data source, will cause the
                data to be appended to the existing records in the data source.  Not
                supported for writing to in-memory files.
                NOTE: append support is limited to specific drivers and GDAL versions.
            dataset_metadata:
                Metadata to be stored at the dataset level in the output file; limited
                to drivers that support writing metadata, such as GPKG, and silently
                ignored otherwise. Keys and values must be strings.
            layer_metadata:
                Metadata to be stored at the layer level in the output file; limited to
                drivers that support writing metadata, such as GPKG, and silently
                ignored otherwise. Keys and values must be strings.
            metadata:
                alias of layer_metadata
            dataset_options:
                Dataset creation options (format specific) passed to OGR. Specify as
                a key-value dictionary.
            layer_options:
                Layer creation options (format specific) passed to OGR. Specify as
                a key-value dictionary.
            **kwargs:
                Additional driver-specific dataset or layer creation options passed
                to OGR. pyogrio will attempt to automatically pass those keywords
                either as dataset or as layer creation option based on the known
                options for the specific driver. Alternatively, you can use the
                explicit `dataset_options` or `layer_options` keywords to manually
                do this (for example if an option exists as both dataset and layer
                option).
        """
        if geometry_name is not None:
            geometry_types = self._df.select(
                geom(geometry_name).st.geometry_type().unique().drop_nulls()
            )
            geometry_type = geometry_types.item() if len(geometry_types) == 1 else "Unknown"

            crs = crs if crs is not None else get_unique_crs_or_raise(self._df, geometry_name)
            geometry = geom(geometry_name).st.to_wkb(output_dimension=4, include_srid=False)
            arrow_obj = self._df.with_columns(geometry).to_arrow()
        else:
            arrow_obj = self._df.to_arrow()
            geometry_type = None  # If geometry_name is None, geometry_type must also be None

        write_arrow(
            arrow_obj,
            path=path,
            layer=layer,
            driver=driver,
            geometry_name=geometry_name,
            geometry_type=geometry_type,
            crs=crs,
            encoding=encoding,
            append=append,
            dataset_metadata=dataset_metadata,
            layer_metadata=layer_metadata,
            metadata=metadata,
            dataset_options=dataset_options,
            layer_options=layer_options,
            **kwargs,
        )

    @overload
    def write_geojson(
        self,
        file: None = None,
        geometry_name: str = "geometry",
    ) -> str: ...

    @overload
    def write_geojson(
        self,
        file: str | Path | IO[bytes] | IO[str],
        geometry_name: str = "geometry",
    ) -> None: ...

    def write_geojson(
        self,
        file: str | Path | IO[bytes] | IO[str] | None = None,
        geometry_name: str = "geometry",
    ) -> str | None:
        r"""Serialize to GeoJSON FeatureCollection representation.

        All geometries in the DataFrame must be of the same type.

        Examples:
            >>> gdf = st.GeoDataFrame({
            ...     "geometry": ["POINT(0 0)", "POINT(1 2)"],
            ...     "name": ["Alice", "Bob"]
            ... })
            >>> geojson = gdf.st.write_geojson()
            >>> print(geojson)
            {"type":"FeatureCollection","features":[{"type":"Feature","geometry":{"type":"Point","coordinates":[0.0,0.0]},"properties":{"name":"Alice"}},{"type":"Feature","geometry":{"type":"Point","coordinates":[1.0,2.0]},"properties":{"name":"Bob"}}]}
            <BLANKLINE>
        """
        geometries = self._df.select(geom(geometry_name).st.to_geojson()).to_series()
        return (
            self._df
            .select(
                type=pl.lit("Feature"),
                geometry=geometries.str.json_decode(),
                properties=pl.struct(cs.exclude(geom(geometry_name)))
                if len(self._df.columns) > 1
                else None,
            )
            .group_by(0)
            .agg(
                type=pl.lit("FeatureCollection"),
                features=pl.struct("type", "geometry", "properties"),
            )
            .select("type", "features")
            .write_ndjson(file)
        )

    @overload
    def write_ndgeojson(
        self,
        file: None = None,
        geometry_name: str = "geometry",
    ) -> str: ...

    @overload
    def write_ndgeojson(
        self,
        file: str | Path | IO[bytes] | IO[str],
        geometry_name: str = "geometry",
    ) -> None: ...

    def write_ndgeojson(
        self,
        file: str | Path | IO[bytes] | IO[str] | None = None,
        geometry_name: str = "geometry",
    ) -> str | None:
        """Serialize to newline-delimited GeoJSON representation.

        All geometries in the DataFrame must be of the same type.

        Examples:
            >>> gdf = st.GeoDataFrame({
            ...     "geometry": ["POINT(0 0)", "POINT(1 2)"],
            ...     "name": ["Alice", "Bob"]
            ... })
            >>> ndgeojson = gdf.st.write_ndgeojson()
            >>> print(ndgeojson)
            {"type":"Feature","geometry":{"type":"Point","coordinates":[0.0,0.0]},"properties":{"name":"Alice"}}
            {"type":"Feature","geometry":{"type":"Point","coordinates":[1.0,2.0]},"properties":{"name":"Bob"}}
            <BLANKLINE>
        """
        geometries = self._df.select(geom(geometry_name).st.to_geojson()).to_series()
        return self._df.select(
            type=pl.lit("Feature"),
            geometry=geometries.str.json_decode(),
            properties=pl.struct(cs.exclude(geom(geometry_name)))
            if len(self._df.columns) > 1
            else None,
        ).write_ndjson(file)

    def plot(self, geometry_name: str = "geometry", **kwargs: Unpack[MarkConfigKwds]) -> alt.Chart:
        """Draw map plot.

        Polars does not implement plotting logic itself but instead defers to
        [`Altair`](https://altair-viz.github.io/).

        `df.st.plot(**kwargs)` is shorthand for
        `alt.Chart({"values": df.st.to_dicts()}).mark_geoshape(**kwargs).interactive()`. Please read
        Altair [GeoShape](https://altair-viz.github.io/user_guide/marks/geoshape.html) documentation
        for available options.

        Examples:
            >>> url = "https://naciscdn.org/naturalearth/110m/cultural/ne_110m_admin_0_countries.zip"
            >>> plot = (
            ...     st.read_file(url)
            ...     .with_columns(st.simplify(tolerance=1))
            ...     .st.plot()
            ...     .encode(color="CONTINENT:N")
            ...     .configure_legend(title=None)
            ...     .properties(height=150)
            ... )

            >>> import altair as alt
            >>> df = st.GeoDataFrame({
            ...     "color": ["red","yellow", "blue"],
            ...     "geometry": [
            ...         "POLYGON((0 0, 0 2, 2 2, 2 0, 0 0))",
            ...         "POLYGON((0 0, 1 2, 2 0, 0 0))",
            ...         "POINT(2 1)"
            ...     ]
            ... })
            >>> plot = (
            ...     df.st.plot(blend="difference")
            ...     .encode(fill=alt.Color("color:N", scale=None))
            ...     .project("identity", reflectY=True, pointRadius=100)
            ...     .properties(height=200)
            ... )
        """
        import altair as alt

        return (
            self._df
            .with_columns(type=pl.lit("Feature"), geometry=geom(geometry_name).st.to_dict())
            .drop([geometry_name] if geometry_name != "geometry" else [])
            .pipe(alt.Chart)
            .mark_geoshape(**kwargs)
            .interactive()
        )

    def explore(
        self,
        geometry_name: str = "geometry",
        *,
        scatterplot_kwargs: ScatterplotLayerKwargs | None = None,
        path_kwargs: PathLayerKwargs | None = None,
        polygon_kwargs: PolygonLayerKwargs | None = None,
        map_kwargs: MapKwargs | None = None,
    ) -> Map:
        from lonboard import viz

        table = (
            self._df
            .with_columns(geometry=geometry_name)
            .drop([geometry_name] if geometry_name != "geometry" else [])
            .pipe(lambda gdf: gdf.to_arrow())
        )

        geom_metadata = {b"ARROW:extension:name": b"geoarrow.wkb"}
        if (crs := get_unique_crs_or_raise(self._df, geometry_name)) is not None:
            crs_metadata = {"crs_type": "wkt2:2019", "crs": crs}
            geom_metadata[b"ARROW:extension:metadata"] = json.dumps(crs_metadata).encode("utf-8")

        geom_field_index = table.schema.get_field_index("geometry")
        geom_field = table.schema.field(geom_field_index).with_metadata(geom_metadata)
        table = table.set_column(geom_field_index, geom_field, table.column(geom_field_index))

        return viz(
            table,
            scatterplot_kwargs=scatterplot_kwargs,
            path_kwargs=path_kwargs,
            polygon_kwargs=polygon_kwargs,
            map_kwargs=map_kwargs,
        )


def get_unique_crs_or_raise(df: GeoDataFrame, geometry_name: str) -> str | None:
    srids = df.select(geom(geometry_name).st.srid()).unique().drop_nulls()
    if len(srids) == 1 and (srid := srids[0, 0]) != 0:
        crs = get_crs_from_code(srid)
        if crs is None:
            msg = f"Couldn't find CRS information for SRID {srid}"
            raise ValueError(msg)
    elif len(srids) > 1:
        msg = "DataFrame with mixed SRIDs aren't supported for this operation"
        raise ValueError(msg)
    else:
        crs = None
    return crs



import geopandas as gpd
import pytest
import shapely

import polars_st as st


def test_to_geopandas_valid_conversion():
    geometry = ["POINT (0 0)", "POINT (1 2)", None]
    name = ["A", "B", "C"]
    gdf = st.GeoDataFrame({
        "geometry": st.GeoSeries(geometry).st.set_srid(4326),
        "name": name,
    })

    geopandas_gdf = gdf.st.to_geopandas()

    assert isinstance(geopandas_gdf, gpd.GeoDataFrame)
    assert list(geopandas_gdf["name"]) == name
    assert list(geopandas_gdf.columns) == ["geometry", "name"]
    assert list(shapely.to_wkt(list(geopandas_gdf.geometry))) == geometry
    assert geopandas_gdf.crs == "EPSG:4326"


def test_to_geopandas_no_srid():
    gdf = st.GeoDataFrame(["POINT (0 0)", "POINT (1 2)", None])
    geopandas_gdf = gdf.st.to_geopandas()
    assert geopandas_gdf.crs is None


def test_to_geopandas_named():
    gdf = st.GeoDataFrame({"geom": ["POINT (0 0)", "POINT (1 2)", None]}, geometry_name="geom")
    geopandas_gdf = gdf.st.to_geopandas(geometry_name="geom")
    assert geopandas_gdf.geometry.name == "geom"


def test_to_geopandas_empty_dataframe():
    gdf = st.GeoDataFrame({"geometry": []})
    geopandas_gdf = gdf.st.to_geopandas()

    assert isinstance(geopandas_gdf, gpd.GeoDataFrame)
    assert geopandas_gdf.empty


def test_to_geopandas_mixed_srids():
    gdf = st.GeoDataFrame({
        "geometry": ["POINT(0 0)", "POINT(1 2)"],
        "srid": [4326, 3857],
    }).with_columns(st.set_srid(srid="srid"))
    msg = "DataFrame with mixed SRIDs aren't supported for this operation"
    with pytest.raises(ValueError, match=msg):
        gdf.pipe(st.st).to_geopandas()


def test_series_to_geopandas():
    gs = st.GeoSeries(["POINT (0 0)", "POINT (1 2)", None])
    geopandas_gs = gs.st.to_geopandas()
    assert isinstance(geopandas_gs, gpd.GeoSeries)


def test_series_to_geopandas_named():
    gs = st.GeoSeries("geom", ["POINT (0 0)", "POINT (1 2)", None])
    geopandas_gs = gs.st.to_geopandas()
    assert isinstance(geopandas_gs, gpd.GeoSeries)
    assert geopandas_gs.geometry.name == "geom"



# ruff: noqa: E501

import math
import warnings
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any, Concatenate, ParamSpec

import numpy as np
import polars as pl
import pytest

import polars_st as st
from polars_st.geoexpr import GeoExprNameSpace as Geo
from polars_st.geometry import (
    GeometryType,
    PolarsCoordinateType,
    PolarsDimensionType,
    PolarsGeometryType,
)

empty_frame = pl.Series("geometry", [], pl.Binary()).to_frame()
none_frame = pl.Series("geometry", [None], pl.Binary()).to_frame()

point_empty = st.GeoDataFrame(["POINT EMPTY"])
point_2d = st.GeoDataFrame(["POINT (1 2)"])
point_3d = st.GeoDataFrame(["POINT (1 2 3)"])
line_empty = st.GeoDataFrame(["LINESTRING EMPTY"])
line_2d = st.GeoDataFrame(["LINESTRING (0 0, 1 1)"])
line_3d = st.GeoDataFrame(["LINESTRING Z (0 0 0, 1 1 1, 2 2 2)"])
poly_empty = st.GeoDataFrame(["POLYGON EMPTY"])
poly_2d = st.GeoDataFrame(["POLYGON ((0 0, 1 0, 1 1, 0 1, 0 0))"])
poly_3d = st.GeoDataFrame(["POLYGON Z ((0 0 1, 1 0 0, 1 1 1, 0 1 0, 0 0 1))"])
multipoint_empty = st.GeoDataFrame(["MULTIPOINT EMPTY"])
multipoint_2d = st.GeoDataFrame(["MULTIPOINT ((0 0), (1 1))"])
multipoint_3d = st.GeoDataFrame(["MULTIPOINT Z ((0 0 0), (1 1 1))"])
multiline_empty = st.GeoDataFrame(["MULTILINESTRING EMPTY"])
multiline_2d = st.GeoDataFrame(["MULTILINESTRING ((0 0, 1 1), (2 2, 3 3))"])
multiline_3d = st.GeoDataFrame(["MULTILINESTRING Z ((0 0 0, 1 1 1), (2 2 2, 3 3 3))"])
multipoly_empty = st.GeoDataFrame(["MULTIPOLYGON EMPTY"])
multipoly_2d = st.GeoDataFrame(["MULTIPOLYGON (((0 0, 1 0, 0 1, 0 0)), ((2 2, 3 2, 2 3, 2 2)))"])
multipoly_3d = st.GeoDataFrame([
    "MULTIPOLYGON Z (((0 0 0, 1 0 0, 0 1 1, 0 0 0)), ((2 2 2, 3 2 3, 2 3 2, 2 2 2)))",
])
collection_empty = st.GeoDataFrame(["GEOMETRYCOLLECTION EMPTY"])
collection_2d = st.GeoDataFrame(["GEOMETRYCOLLECTION (POINT (0 0), LINESTRING (0 0, 1 1))"])
collection_3d = st.GeoDataFrame([
    "GEOMETRYCOLLECTION (POINT (0 0), LINESTRING (0 0, 1 1), POLYGON ((0 0, 1 0, 1 1, 0 0)))",
])
collection_mixed = st.GeoDataFrame([
    "GEOMETRYCOLLECTION (POINT Z (0 0 0), LINESTRING (0 0, 1 1), POLYGON ((0 0, 1 0, 1 1, 0 0)))",
])

base_types = [
    point_empty,
    point_2d,
    point_3d,
    line_empty,
    line_2d,
    line_3d,
    poly_empty,
    poly_2d,
    poly_3d,
    multipoint_empty,
    multipoint_2d,
    multipoint_3d,
    multiline_empty,
    multiline_2d,
    multiline_3d,
    multipoly_empty,
    multipoly_2d,
    multipoly_3d,
    collection_empty,
    collection_2d,
    collection_3d,
    collection_mixed,
]

dummy_point = point_2d.item()
dummy_line = line_2d.item()

P = ParamSpec("P")
FunctionCallable = Callable[Concatenate[Geo, P], pl.Expr]


@dataclass()
class Function:
    call: FunctionCallable
    dtype: pl.DataType
    args: dict[str, Any] = field(default_factory=dict)

    def __call__(self):
        return self.call(st.geom().st, **self.args)


functions = [
    Function(Geo.geometry_type, PolarsGeometryType),
    Function(Geo.coordinate_type, PolarsCoordinateType),
    Function(Geo.coordinate_dimension, pl.UInt8()),
    Function(Geo.dimension, PolarsDimensionType),
    Function(Geo.srid, pl.Int32()),
    Function(Geo.set_srid, pl.Binary(), {"srid": 3857}),
    Function(Geo.to_srid, pl.Binary(), {"srid": 3857}),
    Function(Geo.x, pl.Float64()),
    Function(Geo.y, pl.Float64()),
    Function(Geo.z, pl.Float64()),
    Function(Geo.m, pl.Float64()),
    Function(Geo.exterior_ring, pl.Binary()),
    Function(Geo.count_points, pl.UInt32()),
    Function(Geo.count_interior_rings, pl.UInt32()),
    Function(Geo.count_geometries, pl.UInt32()),
    Function(Geo.get_point, pl.Binary(), {"index": 0}),
    Function(Geo.get_interior_ring, pl.Binary(), {"index": 0}),
    Function(Geo.get_geometry, pl.Binary(), {"index": 0}),
    Function(Geo.parts, pl.List(pl.Binary())),
    Function(Geo.interior_rings, pl.List(pl.Binary())),
    Function(Geo.precision, pl.Float64()),
    Function(Geo.set_precision, pl.Binary(), {"grid_size": 1.0, "mode": "valid_output"}),
    Function(Geo.set_precision, pl.Binary(), {"grid_size": 1.0, "mode": "no_topo"}),
    Function(Geo.set_precision, pl.Binary(), {"grid_size": 1.0, "mode": "keep_collapsed"}),
    Function(Geo.to_wkt, pl.String()),
    Function(Geo.to_ewkt, pl.String()),
    Function(Geo.to_wkb, pl.Binary()),
    Function(Geo.to_geojson, pl.String()),
    Function(Geo.to_dict, pl.Object()),
    Function(Geo.to_shapely, pl.Object()),
    Function(Geo.area, pl.Float64()),
    Function(Geo.bounds, pl.Array(pl.Float64, 4)),
    Function(Geo.length, pl.Float64()),
    Function(Geo.distance, pl.Float64(), {"other": dummy_point}),
    Function(Geo.hausdorff_distance, pl.Float64(), {"other": dummy_point, "densify": None}),
    Function(Geo.hausdorff_distance, pl.Float64(), {"other": dummy_point, "densify": 0.5}),
    Function(Geo.frechet_distance, pl.Float64(), {"other": dummy_point, "densify": None}),
    Function(Geo.frechet_distance, pl.Float64(), {"other": dummy_point, "densify": 0.5}),
    Function(Geo.minimum_clearance, pl.Float64()),
    Function(Geo.has_z, pl.Boolean()),
    Function(Geo.has_m, pl.Boolean()),
    Function(Geo.is_ccw, pl.Boolean()),
    Function(Geo.is_closed, pl.Boolean()),
    Function(Geo.is_empty, pl.Boolean()),
    Function(Geo.is_ring, pl.Boolean()),
    Function(Geo.is_simple, pl.Boolean()),
    Function(Geo.is_valid, pl.Boolean()),
    Function(Geo.is_valid_reason, pl.String()),
    Function(Geo.crosses, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.contains, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.contains_properly, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.covered_by, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.covers, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.disjoint, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.dwithin, pl.Boolean(), {"other": dummy_point, "distance": 1.0}),
    Function(Geo.intersects, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.overlaps, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.touches, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.within, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.equals, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.equals_exact, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.equals_identical, pl.Boolean(), {"other": dummy_point}),
    Function(Geo.relate, pl.String(), {"other": dummy_point}),
    Function(Geo.relate_pattern, pl.Boolean(), {"other": dummy_point, "pattern": "*********"}),
    Function(Geo.difference, pl.Binary(), {"other": dummy_point, "grid_size": None}),
    Function(Geo.difference, pl.Binary(), {"other": dummy_point, "grid_size": 0.5}),
    Function(Geo.intersection, pl.Binary(), {"other": dummy_point, "grid_size": None}),
    Function(Geo.intersection, pl.Binary(), {"other": dummy_point, "grid_size": 0.5}),
    Function(Geo.symmetric_difference, pl.Binary(), {"other": dummy_point, "grid_size": None}),
    Function(Geo.symmetric_difference, pl.Binary(), {"other": dummy_point, "grid_size": 0.5}),
    Function(Geo.union, pl.Binary(), {"other": dummy_point, "grid_size": None}),
    Function(Geo.union, pl.Binary(), {"other": dummy_point, "grid_size": 0.5}),
    Function(Geo.unary_union, pl.Binary(), {"grid_size": None}),
    Function(Geo.unary_union, pl.Binary(), {"grid_size": 0.5}),
    Function(Geo.cast, pl.Binary(), {"into": "GeometryCollection"}),
    Function(Geo.multi, pl.Binary()),
    Function(Geo.boundary, pl.Binary()),
    Function(Geo.coverage_union, pl.Binary()),
    Function(Geo.buffer, pl.Binary(), {"distance": 1.0}),
    Function(Geo.offset_curve, pl.Binary(), {"distance": 1.0}),
    Function(Geo.centroid, pl.Binary()),
    Function(Geo.center, pl.Binary()),
    Function(Geo.clip_by_rect, pl.Binary(), {"bounds": [0.0, 0.0, 1.0, 1.0]}),
    Function(Geo.concave_hull, pl.Binary()),
    Function(Geo.convex_hull, pl.Binary()),
    Function(Geo.segmentize, pl.Binary(), {"max_segment_length": 1.0}),
    Function(Geo.envelope, pl.Binary()),
    Function(Geo.extract_unique_points, pl.Binary()),
    Function(Geo.build_area, pl.Binary()),
    Function(Geo.make_valid, pl.Binary()),
    Function(Geo.normalize, pl.Binary()),
    Function(Geo.node, pl.Binary()),
    Function(Geo.point_on_surface, pl.Binary()),
    Function(Geo.remove_repeated_points, pl.Binary()),
    Function(Geo.reverse, pl.Binary()),
    Function(Geo.snap, pl.Binary(), {"other": dummy_point, "tolerance": 1.0}),
    Function(Geo.simplify, pl.Binary(), {"tolerance": 1.0, "preserve_topology": False}),
    Function(Geo.simplify, pl.Binary(), {"tolerance": 1.0, "preserve_topology": True}),
    Function(Geo.flip_coordinates, pl.Binary()),
    Function(Geo.minimum_rotated_rectangle, pl.Binary()),
    Function(Geo.maximum_inscribed_circle, pl.Binary(), {"tolerance": 1.0}),
    Function(Geo.translate, pl.Binary()),
    Function(Geo.rotate, pl.Binary(), {"angle": 90}),
    Function(Geo.scale, pl.Binary()),
    Function(Geo.skew, pl.Binary()),
    Function(Geo.interpolate, pl.Binary(), {"distance": 1.0, "normalized": False}),
    Function(Geo.interpolate, pl.Binary(), {"distance": 1.0, "normalized": True}),
    Function(Geo.project, pl.Float64(), {"other": dummy_point, "normalized": False}),
    Function(Geo.project, pl.Float64(), {"other": dummy_point, "normalized": True}),
    Function(Geo.substring, pl.Binary(), {"start": 0.0, "end": 0.0}),
    Function(Geo.line_merge, pl.Binary(), {"directed": True}),
    Function(Geo.line_merge, pl.Binary(), {"directed": False}),
    Function(Geo.shared_paths, pl.Binary(), {"other": dummy_line}),
    Function(Geo.shortest_line, pl.Binary(), {"other": dummy_point}),
    Function(Geo.count_coordinates, pl.UInt32()),
    Function(Geo.coordinates, pl.List(pl.List(pl.Float64))),
]


@dataclass()
class Aggregate:
    call: FunctionCallable
    dtype: pl.DataType
    identity: Any


aggregates = [
    Aggregate(Geo.voronoi_polygons, pl.Binary(), collection_empty.item()),
    Aggregate(Geo.delaunay_triangles, pl.Binary(), collection_empty.item()),
    Aggregate(Geo.polygonize, pl.Binary(), collection_empty.item()),
    Aggregate(Geo.total_bounds, pl.Array(pl.Float64(), 4), np.full(4, np.nan)),
    Aggregate(Geo.intersection_all, pl.Binary(), collection_empty.item()),
    Aggregate(Geo.symmetric_difference_all, pl.Binary(), collection_empty.item()),
    Aggregate(Geo.union_all, pl.Binary(), collection_empty.item()),
    Aggregate(Geo.coverage_union_all, pl.Binary(), collection_empty.item()),
    Aggregate(Geo.collect, pl.Binary(), collection_empty.item()),
]


@pytest.mark.parametrize("frame", [empty_frame])
@pytest.mark.parametrize("func", functions)
def test_functions_empty_frame(frame: pl.DataFrame, func: Function):
    """Functions should work on empty frames."""
    result = frame.select(func())
    assert result.schema == pl.Schema([("geometry", func.dtype)])
    assert len(result) == 0


@pytest.mark.parametrize("frame", [none_frame])
@pytest.mark.parametrize("func", functions)
def test_functions_none_frame(frame: pl.DataFrame, func: Function):
    """Functions should work on full-null frames."""
    result = frame.select(func())
    assert result.schema == pl.Schema([("geometry", func.dtype)])
    assert len(result) == 1
    assert result.item() is None


@pytest.mark.parametrize("frame", [empty_frame])
@pytest.mark.parametrize("func", functions)
def test_functions_empty_frame_agg(frame: pl.DataFrame, func: Function):
    """Functions should work on empty frames in aggregation context."""
    # Should file a bug report in polars for that (cannot concatenate empty list of arrays)
    if func.call == Geo.bounds:
        return
    result = frame.group_by(0).agg(func()).drop("literal")
    assert result.schema == pl.Schema([("geometry", pl.List(func.dtype))])
    assert len(result) == 0


@pytest.mark.parametrize("frame", [none_frame])
@pytest.mark.parametrize("func", functions)
def test_functions_none_frame_agg(frame: pl.DataFrame, func: Function):
    """Functions should work on full-null frames in aggregation context."""
    # Skip since List(Object) is not supported
    if func.call in {Geo.to_dict, Geo.to_shapely}:
        return
    result = frame.group_by(0).agg(func()).drop("literal")
    assert result.schema == pl.Schema([("geometry", pl.List(func.dtype))])
    assert len(result) == 1
    assert result.get_column("geometry").list.len().item() == 1
    assert result.get_column("geometry").list.get(0).item() is None


@pytest.mark.parametrize("frame", [empty_frame.group_by(0).agg(st.geom())])
@pytest.mark.parametrize("func", functions)
def test_functions_empty_list_frame(frame: pl.DataFrame, func: Function):
    """Functions should work on empty lists."""
    # Skip since List(Object) is not supported
    if func.call in {Geo.to_dict, Geo.to_shapely}:
        return
    result = frame.select(st.geom().list.eval(func.call(st.element().st, **func.args)))
    assert result.schema == pl.Schema([("geometry", pl.List(func.dtype))])
    assert len(result) == 0


@pytest.mark.parametrize("frame", [none_frame.group_by(0).agg(st.geom())])
@pytest.mark.parametrize("func", functions)
def test_functions_none_list_frame(frame: pl.DataFrame, func: Function):
    """Functions should work on full-null lists."""
    # Skip since List(Object) is not supported
    if func.call in {Geo.to_dict, Geo.to_shapely}:
        return
    result = frame.select(st.geom().list.eval(func.call(st.element().st, **func.args)))
    assert result.schema == pl.Schema([("geometry", pl.List(func.dtype))])
    assert result.item().item() is None


@pytest.mark.parametrize("frame", [none_frame, empty_frame])
@pytest.mark.parametrize("func", aggregates)
def test_aggregates(frame: pl.DataFrame, func: Aggregate):
    """Aggregations should work on empty and full-null frames."""
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="invalid value encountered in coverage_union")
        result = frame.select(func.call(st.geom().st))
    assert result.schema == pl.Schema([("geometry", func.dtype)])
    assert len(result) == 1
    item = result.item()
    if isinstance(item, pl.Series):
        # item() returns a series if dtype is List or Array (for st.total_bounds)
        assert np.array_equal(item.to_numpy(), func.identity, equal_nan=True)
    else:
        assert item == func.identity


@pytest.mark.parametrize("frame", base_types)
@pytest.mark.parametrize("func", functions)
def test_functions_all_types_frame(frame: pl.DataFrame, func: Function):  # noqa: C901, PLR0912, PLR0915
    """Functions should work on every geometry type."""
    geom_type: GeometryType = frame.select(st.geometry_type()).item()
    geom_empty: bool = frame.select(st.is_empty()).item()
    error = None

    if (
        func.call
        in {
            Geo.difference,
            Geo.symmetric_difference,
            Geo.union,
            Geo.coverage_union,
        }
        and func.args.get("grid_size", 0) is not None
        and geom_type == "GeometryCollection"
        and not geom_empty
    ):
        error = "IllegalArgumentException: Overlay input is mixed-dimension"

    if func.call == Geo.coverage_union and frame is collection_mixed:
        error = "IllegalArgumentException: Overlay input is mixed-dimension"

    if func.call == Geo.shared_paths and geom_type not in {"LineString", "MultiLineString"}:
        error = "IllegalArgumentException: Geometry is not lineal"

    if func.call == Geo.exterior_ring and geom_type not in {"Polygon", "CurvePolygon"}:
        error = "IllegalArgumentException: Invalid argument \\(must be a Surface\\)"

    if func.call in {
        Geo.count_interior_rings,
        Geo.get_interior_ring,
        Geo.interior_rings,
    } and geom_type not in {"Polygon", "CurvePolygon"}:
        error = "IllegalArgumentException: Argument is not a Surface"

    if func.call in {Geo.x, Geo.y, Geo.z, Geo.m} and geom_type != "Point":
        error = "IllegalArgumentException: Argument is not a Point"

    if func.call == Geo.boundary and geom_type == "GeometryCollection":
        error = "IllegalArgumentException: Operation not supported by GeometryCollection"

    if func.call in {Geo.count_points, Geo.get_point} and geom_type != "LineString":
        error = "IllegalArgumentException: Argument is not a SimpleCurve"

    if func.call == Geo.is_closed and geom_type not in {
        "LineString",
        "MultiLineString",
        "MultiCurve",
    }:
        error = "IllegalArgumentException: Argument is not a Curve, MultiLineString, or MultiCurve"

    if (
        func.call == Geo.substring
        and geom_type not in {"LineString", "MultiLineString"}
        and not (geom_type == "GeometryCollection" and geom_empty)
    ):
        error = "IllegalArgumentException: Input geometry must be linear"

    if (
        func.call == Geo.project
        and geom_type != "LineString"
        and not (geom_type == "MultiLineString" and not geom_empty)
    ):
        error = "IllegalArgumentException: LinearIterator only supports lineal geometry components"
        if geom_type not in {"Point", "Polygon"} and geom_empty:
            geos_func = "GEOSProjectNormalized_r" if func.args["normalized"] else "GEOSProject_r"
            error = f"{geos_func} failed"

    if (
        func.call == Geo.interpolate
        and geom_type not in {"LineString", "MultiLineString"}
        and not (geom_type in {"MultiPoint", "MultiPolygon", "GeometryCollection"} and geom_empty)
        and not (func.args["normalized"] and geom_empty)
    ):
        error = "IllegalArgumentException: LinearIterator only supports lineal geometry components"
        if func.args["normalized"] and geom_type in {"Point", "MultiPoint"} and not geom_empty:
            error = "IllegalArgumentException: LinearLocation::getCoordinate only works with LineString geometries"

    if func.call == Geo.coverage_union and geom_type not in {
        "MultiPoint",
        "MultiLineString",
        "MultiPolygon",
        "GeometryCollection",
        "CompoundCurve",
        "MultiCurve",
        "MultiSurface",
    }:
        error = "Geometry must be a collection"

    if func.call == Geo.maximum_inscribed_circle:
        if geom_type not in {"Polygon", "MultiPolygon"}:
            if geom_type == "GeometryCollection":
                error = "IllegalArgumentException: Operation not supported by GeometryCollection"
            else:
                error = "IllegalArgumentException: Input must be a Polygon or MultiPolygon"
        elif geom_empty:
            error = "IllegalArgumentException: Empty input is not supported"

    if func.call == Geo.to_srid:
        frame = frame.select(st.geom().st.set_srid(4326))

    if error is not None:
        with pytest.raises(pl.exceptions.ComputeError, match=f"{error}"):
            frame.select(func())
        return

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="invalid value encountered in voronoi_polygons")
        result = frame.select(func())

    assert result.schema == pl.Schema([("geometry", func.dtype)])


def test_sjoin_dwithin_point_fast_path():
    """Sjoin dwithin must match when the right geometry is a Point (#66)."""
    left = st.GeoDataFrame(["POLYGON ((0 0, 90 0, 90 90, 0 90, 0 0))"])
    right = st.GeoDataFrame(["POINT (-5 -5)"])
    result = left.st.sjoin(right, predicate="dwithin", distance=math.hypot(5, 5))
    assert len(result) == 1
    result = left.st.sjoin(right, predicate="dwithin", distance=math.hypot(5, 5) - 0.1)
    assert len(result) == 0



import inspect
from typing import TYPE_CHECKING

import polars_st as st

if TYPE_CHECKING:
    from collections.abc import Callable


def signature_matches(left: inspect.Signature, right: inspect.Signature) -> bool:
    """Returns True if both signatures share the same parameters names and default values."""
    assert left.return_annotation == right.return_annotation.replace("Series", "Expr")

    assert set(left.parameters) == set(right.parameters)

    for a, b in zip(
        left.parameters.values(),
        right.parameters.values(),
        strict=True,
    ):
        assert a.name == b.name
        assert a.default == b.default
        assert str(a.annotation) == str(b.annotation)
    return True


def patch_sugar_signature(sig: inspect.Signature) -> inspect.Signature:
    """Returns a sugar function signature with patched parameters for comparison with GeoExpr."""
    params = [
        p.replace(name="self", annotation=inspect._empty) if i == 0 else p  # noqa: SLF001
        for i, p in enumerate(sig.parameters.values())
    ]
    return sig.replace(parameters=params)


def test_series_signatures_matches_expr():
    """All GeoExpr functions should have a matching GeoSeries function."""
    expr_funcs: dict[str, Callable] = {
        name: func
        for name, func in st.geoexpr.GeoExprNameSpace.__dict__.items()
        if callable(func) and not name.startswith("_")
    }
    series_funcs: dict[str, Callable] = {
        name: func
        for name, func in st.geoseries.GeoSeriesNameSpace.__dict__.items()
        if callable(func)
        and not name.startswith("_")
        and name not in {"plot", "explore", "to_geopandas"}
    }
    assert set(expr_funcs.keys()) == set(series_funcs.keys())
    for expr_func, series_func in zip(expr_funcs.values(), series_funcs.values(), strict=True):
        assert expr_func.__name__ == series_func.__name__
        assert signature_matches(
            inspect.signature(expr_func),
            inspect.signature(series_func),
        ), expr_func.__name__


def test_sugar_signatures_matches_expr():
    """All GeoExpr functions should have a matching GeoSeries function."""
    expr_funcs: dict[str, Callable] = {
        name: func
        for name, func in st.geoexpr.GeoExprNameSpace.__dict__.items()
        if callable(func) and not name.startswith("_")
    }
    sugar_funcs: dict[str, Callable] = {
        name: func
        for name, func in st.sugar.__dict__.items()
        if callable(func) and name in st.sugar.__all__
    }
    expr_funcs = {name: func for name, func in expr_funcs.items() if name in sugar_funcs}

    assert set(sugar_funcs.keys()).issubset(expr_funcs.keys())
    for expr_func, sugar_func in zip(expr_funcs.values(), sugar_funcs.values(), strict=True):
        assert expr_func.__name__ == sugar_func.__name__
        assert signature_matches(
            inspect.signature(expr_func),
            patch_sugar_signature(inspect.signature(sugar_func)),
        )



from collections.abc import Callable


def is_empty_method(func: Callable) -> bool:
    """Return True if the function body is only '...'."""

    def _empty_with_docstring() -> None:
        """Docstring."""
        ...

    def _empty_without_docstring() -> None: ...

    empty_bytecodes = {
        _empty_with_docstring.__code__.co_code,
        _empty_without_docstring.__code__.co_code,
    }

    return func.__code__.co_code in empty_bytecodes
